/* Virtuino SE Json example: Overview
 * Supported boards: Arduino uno + Ethernet shield
 * Created by Ilias Lamprou
 * Updated Feb 27 2019
 * More examples: https://virtuino.com/index.php/virtuino-se/20-virtuino-se-code-examples
 */
// ------ VirtuinoSE V Memory Table
//  V0 memory -> digital output pin 6  connect a led or relay to this pin
//  V1 memory -> digital input pin 7   connect a button to this pin
//  V2 memory -> variable 
//  V3 memory -> Sensor1 
//  V3 memory -> Sensor2 
//  V4 memory -> A0  analog input
//  V5 memory -> reserved
//  V6 memory -> reserved
//  ...
//  V15 memory -> reserved

#define outPin_1 6
#define inPin_1 7


#include <SPI.h>
#include <Ethernet.h>
//#include <Ethernet2.h>    // enable this line to support Ethernet 2
 
byte mac[] = {0xDE, 0xAD, 0xBE, 0xEF, 0xFE, 0xED};       // Set the ethernet shield mac address.
IPAddress ip(192, 168, 1, 150);                          // Set the ethernet shield ip address. Check your gateway ip address first
EthernetServer server(8000);                             // Server port =8000
    

//------------------- USER SETTINGS ---------------------------------
#define VIRTUINO_KEY  "1234"      // Virtuino KEY, only the requests that include this key will be acceptable. // disable the line if you don't want to use a key                          

#define V_memory_count 16         // the size of V memory. You can change it to a number <255)
float V_memory[V_memory_count];    // This array is synchronized with Virtuino V memory. You can change the type to int, long etc.


boolean debug = true;              // set this variable to false on the finale code to decrease the request time

//--- VARIABLES
float   my_Variable = 0;
byte inputPinState=0;

                                  
unsigned long timeStrored=0;

//============================================================== setup
//==============================================================
void setup() {
  if (debug) {
    Serial.begin(9600);
    while (!Serial) continue;
  }
  Ethernet.begin(mac, ip);
  server.begin();
  if (debug) Serial.println(Ethernet.localIP());

  pinMode(outPin_1, OUTPUT);            
  pinMode(inPin_1, INPUT);
  
  }




//============================================================== loop
//==============================================================
void loop() {
  virtuinoRun();        //Necessary function to communicate with Virtuino. Client handler

  //--- How to control outputs from Virtuino 
    if (V_memory[0]==1) digitalWrite(outPin_1,HIGH);      // On Virtuino panel add switch to control the pin outPin_13
    else digitalWrite(outPin_1,LOW);
  
  
  //--- How to send to Virtuino the state of a pin every time it is changed 
  byte pinState = digitalRead(inPin_1);    // read the state of pin inPin_1
  if (pinState!=V_memory[1]) {
      V_memory[1] = pinState;        // Store the pin state to V1. On Virtuino panel add a led to virtual memory V1  
  }


  //--- How to read a variable from Virtuino
  if (my_Variable!=V_memory[2]) {        // Read the V2 memory.  On Virtuino panel add a regulator  to virtual memory V2  
      my_Variable==V_memory[21];
  }
  
  //--- How to send to Virtuino sensor values
   if (millis()-timeStrored>5000) {      // read sensor value every 5 seconds. Avoid to use delay on your code
      V_memory[3]= random(100);           // copy the sensor value to V5
      V_memory[4] = random(100);          // copy the sensor value to V6
      V_memory[5] = analogRead(A0);       // copy the A0 value V7
      timeStrored=millis();               // On Virtuino panel add value displays or instrument  to virtual memories V3,V4,V5
  }


}


//============================================================== onCommandReceived
//==============================================================
/* This function is called every time Virtuino app sends a request to server 
 * The 'variableIndex' is the V memory index of Virtuino app
 * The 'valueAsText' is the value that has sent from the app
 * If the 'valueAsText" is ? you have to return the value that you want to be displayed on Virtuino panel
 */
 String onCommandReceived(int variableIndex, String valueAsJson){     
   String valueAsText = getJsonValue(valueAsJson,"value");    // read the value from the received Json string 
   if (debug) Serial.println("New value received: V"+String(variableIndex)+"  value="+valueAsText);

  
  if (valueAsText=="?") return  String(V_memory[variableIndex]);   // return the value of the arduino V memory array
  else {
    float value = valueAsText.toFloat();        // convert the value to float. The valueAsText have to be numerical
    V_memory[variableIndex]=value;              // copy the received value to arduino V memory array
  }
  
  return "";
}















//================== Virtuino Library (included in sketch) ===============
//==============================================================
//==============================================================
/* Only the next functions are nedded to communicate with Virtuino 
 * Avoid to make changes to the code below  
 */


#define WRONG_KEY_MESSAGE  "Wrong Key"  
#define WELLCOME_MESSAGE  "Hello Virtuino"  


//============================================================== virtuinoRun
//==============================================================
  void virtuinoRun(){
  EthernetClient client = server.available();
  if (client) {
    if (debug) Serial.println("Client connected");
    String data="";
    while (client.connected()) {
      while (client.available()>0) {
        char c = client.read();
        data+=c;
        if (debug) Serial.write(c);
      }
     if (debug) Serial.println("\nReceived data: "+data);
     String response= getResponseAsJson(data);
     if (debug) Serial.println("Response : "+response);
     client.print(response);
     client.stop();
    }
    if (debug) Serial.println("Client disconnected");
  }
 }



//============================================================== prepareResponse
//==============================================================
String getResponseAsJson(String data){
    //---check the KEY
   #ifdef VIRTUINO_KEY 
      String key = getJsonValue(data, "key");
      if (debug) Serial.println("Key="+key);
      if (!key.equals(VIRTUINO_KEY)) return "{\"status\":\"-1\",\"message\":\""+String(WRONG_KEY_MESSAGE)+"\"}";
   #endif

    //---check connection status
    String statusInfo = getJsonValue(data, "status");
    if (debug) Serial.println("statusInfo="+statusInfo);
    if (statusInfo.equals("0")) return "{\"status\":\"2\",\"message\":\""+String(WELLCOME_MESSAGE)+"\"}";
  
   String response="{\"status\":\"1\"";
   for (byte i=0;i<V_memory_count;i++){
      // easy way to read a Json object without installing extra json libraries
      String jsonTag="V"+String(i)+"_";
      int pos = data.indexOf(jsonTag);
      if (pos>0) {
        int startPos= data.indexOf("{",pos+1);
        if (startPos>pos) {
          int endPos= data.indexOf("}", startPos+2);
          if (endPos>startPos) {
            String variableAsJson = data.substring(startPos,endPos+1);
            String variableValue=onCommandReceived(i,variableAsJson);
           if (variableValue.length()>0) response+=",\""+jsonTag+"\":{\"value\":\""+urlencode(&variableValue)+"\"}";
          }
        }
      }
      
   } //for
   response+="}";
  return response;
}

//==============================================================
// easy way to read a Json text value without installing extra json libraries
  String getJsonValue(String jsonAsText, String jsonTag){
    int pos = jsonAsText.indexOf(jsonTag);
    if (pos>0) {
      int startPos=jsonAsText.indexOf(":\"",pos);
      int lastPos = jsonAsText.indexOf("\"",startPos+2);
      String value=jsonAsText.substring(startPos+2,lastPos);
      value=urldecode(&value);
      if (lastPos>startPos) return value;
    }
    return "";
 }


  //=================== ENCODE - DECODE Library (included in sketch) ==========
  //=================================================================
  
  //================================================================ urldecode
String urldecode(String* str){
   String encodedString="";
    char c;
    char code0;
    char code1;
    for (int i =0; i < str->length(); i++){
        c=str->charAt(i);
      if (c == '+'){
        encodedString+=' ';  
      }else if (c == '%') {
        i++;
        code0=str->charAt(i);
        i++;
        code1=str->charAt(i);
        c = (h2int(code0) << 4) | h2int(code1);
        encodedString+=c;
      } else{
         encodedString+=c;  
      }
      yield();
    }
    
   return encodedString;
}
//================================================================ urlencode
String urlencode(String* str){
   String encodedString="";
    char c;
    char code0;
    char code1;
    char code2;
    for (int i =0; i < str->length(); i++){
      c=str->charAt(i);
      if (c == ' '){
        encodedString+= '+';
      } else if (isalnum(c)){
        encodedString+=c;
      } else{
        code1=(c & 0xf)+'0';
        if ((c & 0xf) >9){
            code1=(c & 0xf) - 10 + 'A';
        }
        c=(c>>4)&0xf;
        code0=c+'0';
        if (c > 9){
            code0=c - 10 + 'A';
        }
        code2='\0';
        encodedString+='%';
        encodedString+=code0;
        encodedString+=code1;
        //encodedString+=code2;
      }
      yield();
    }
    return encodedString;
}
//================================================================ h2int
unsigned char h2int(char c){
    if (c >= '0' && c <='9'){
        return((unsigned char)c - '0');
    }
    if (c >= 'a' && c <='f'){
        return((unsigned char)c - 'a' + 10);
    }
    if (c >= 'A' && c <='F'){
        return((unsigned char)c - 'A' + 10);
    }
    return(0);
}



